/**
 * @typedef {(message:Data)=>void} MessageCallback
 * @template Data
 */
/**
 * Class representing any connection that can be used to send data back and forth, like a WebSocket,
 * an RTCDataChannel or even just a plain object used to communicate between two scripts.
 * @template SendData data that is send from this endpoint
 * @template ReceiveData data that is send to this endpoint
 */
export default class DataChannel {
	/**
	 * Constructs a new DataChannel endpoint from the given callbacks once the promise resolves.
	 * @param {()=>Promise<{send:(data:SendData)=>void,onMessage:(callback:MessageCallback<ReceiveData>)=>void,close?:()=>void}>} callback
	 */
	constructor(callback){
		/** @type {MessageCallback<ReceiveData>[]} */
		this._onMessageCallbacks = [];
		/** @type {Promise<void>} */
		this._readyPromise = new Promise(async resolve=>{
			// halts execution of this and puts in the microtask queue so the constructor finishes execution first, before the callback is called,
			// because otherwise the callback can't access the this object without throwing errors.
			// timing is difficult, okay? I'm sorry
			await Promise.resolve();
			let c = await callback();
			this._send = c.send;
			c.onMessage(message=>{
				this._onMessageCallbacks.slice().forEach(callback=>callback(message));
			});
			this._close = c.close||(()=>{});
			resolve();
		});
	}

	/**
	 * Waits until the connection is ready.
	 */
	async waitUntilReady(){
		return this._readyPromise;
	}

	/**
	 * Sends the given message to the other endpoint of the connection.
	 * 
	 * If the connection isn't ready already, waits until it is before sending the message and then returns.
	 * @param {SendData} message
	 */
	async send(message){
		await this._readyPromise;
		this._send(message);
	}

	/**
	 * Registers a callback to be called whenever a message is received.
	 * @param {MessageCallback<ReceiveData>} callback
	 */
	onMessage(callback){
		this._onMessageCallbacks.push(callback);
	}

	/**
	 * Removes a callback registered using `onMessage(callback)`.
	 * @param {MessageCallback<ReceiveData>} callback
	 */
	removeMessageCallback(callback){
		let index = this._onMessageCallbacks.indexOf(callback);
		if (index!==-1){
			this._onMessageCallbacks.splice(index,1);
		}
	}

	/**
	 * Returns a promise that eventually resolves to the next message received.
	 * @returns {Promise<ReceiveData>}
	 */
	next(){
		return new Promise(resolve=>{
			/** @type {(message:ReceiveData)=>void} */
			let callback = message=>{
				this.removeMessageCallback(callback);
				resolve(message);
			}
			this.onMessage(callback);
		});
	}

	/**
	 * Closes the connection.
	 */
	close(){
		this._close();
	}

	/**
	 * Creates a new DataChannel that send messages via an already-existing DataChannel.
	 * @template OldSendData
	 * @template OldReceiveData
	 * @template NewSendData
	 * @template NewReceiveData
	 * @param {DataChannel<OldSendData,OldReceiveData>} dataChannel DataChannel to send messages over
	 * @param {(message:NewSendData)=>OldSendData} modifier modifier to outgoing messages
	 * @param {(message:OldReceiveData)=>NewReceiveData} filter modifier to incoming messages, can act as a filter by returning `null` 
	 */
	static from(dataChannel,modifier,filter){
		return new DataChannel(async()=>{
			await dataChannel.waitUntilReady();
			/** @type {(message:NewSendData)=>void} */
			let send = message=>{
				dataChannel.send(modifier(message));
			};
			/** @type {(callback:(message:NewReceiveData)=>void)=>void} */
			let onMessage = callback=>{
				dataChannel.onMessage(message=>{
					let m = filter(message);
					if (m!==null){
						callback(m);
					}
				});
			};
			return {send,onMessage};
		});
	}

	/**
	 * Creates a pair of DataChannel endpoints talking directly to each other.
	 * 
	 * Whenever one of those sends a message, all onMessage callbacks registered on the other one get queued as a microtask.
	 * @template DataToA
	 * @template DataToB
	 * @returns {[DataChannel<DataToB,DataToA>,DataChannel<DataToA,DataToB>]}
	 */
	static createPair(){
		/** @type {((message:DataToA)=>void)[]} */
		let onMessageCallbacksA = [];
		/** @type {((message:DataToB)=>void)[]} */
		let onMessageCallbacksB = [];
		/** @type {DataChannel<DataToB,DataToA>} */
		let endpointA = new DataChannel(()=>Promise.resolve({send:message=>{
			onMessageCallbacksB.forEach(callback=>queueMicrotask(()=>callback(message)));
		},onMessage:callback=>{
			onMessageCallbacksA.push(callback);
		}}));
		/** @type {DataChannel<DataToA,DataToB>} */
		let endpointB = new DataChannel(()=>Promise.resolve({send:message=>{
			onMessageCallbacksA.forEach(callback=>queueMicrotask(()=>callback(message)));
		},onMessage:callback=>{
			onMessageCallbacksB.push(callback);
		}}));
		return [endpointA,endpointB];
	}
}
/**
 * Class representing a connection to a server via a WebSocket.
 * @extends DataChannel<string,string>
 */
export class SocketDataChannel extends DataChannel {
	/**
	 * Creates a new WebSocket connection to the given url.
	 * @param {string} url
	 */
	constructor(url){
		super(async()=>{
			this._webSocket = new WebSocket(url);
			/** @type {(message:string)=>void} */
			let send = message=>{
				this._webSocket.send(message);
			};
			/** @type {(callback:(message:string)=>void)=>void} */
			let onMessage = callback=>{
				this._webSocket.addEventListener("message",e=>{
					callback(e.data+"");
				});
			};
			let close = ()=>{
				this._webSocket.close();
			}
			await new Promise(resolve=>{
				this._webSocket.onopen = resolve;
			});
			return {send,onMessage,close};
		});
	}
}
/**
 * Class representing a connection to another browser via an `RTCPeerConnection` with a single `RTCDataChannel`.
 * 
 * Not responsible for maintaining the connection, only for using it.
 * @extends DataChannel<string,string>
 */
export class PeerDataChannel extends DataChannel {
	/**
	 * Creates a new PeerDataChannel object wrapping the given `RTCPeerConnection` and `RTCDataChannel`.
	 * @param {RTCPeerConnection} connection
	 * @param {RTCDataChannel} dataChannel
	 */
	constructor(connection,dataChannel){
		super(async()=>{
			this._connection = connection;
			this._dataChannel = dataChannel;
			/** @type {(message:string)=>void} */
			let send = message=>{
				this._dataChannel.send(message);
			};
			/** @type {(callback:(message:string)=>void)=>void} */
			let onMessage = callback=>{
				this._dataChannel.addEventListener("message",e=>{
					callback(e.data+"");
				});
			};
			let close = ()=>{
				this._connection.close();
			}
			/*console.log("Signaling state:",this._connection.signalingState);
			if (this._connection.signalingState!=="stable"){
				await new Promise(resolve=>{
					this._connection.addEventListener("signalingstatechange",e=>{
						console.log("Signaling state changed to:",connection.signalingState);
						if (this._connection.signalingState==="stable"){
							resolve();
						}
					});
				});
			}*/
			console.log("DataChannel state:",this._dataChannel.readyState);
			if (this._dataChannel.readyState!=="open"){
				await new Promise(resolve=>{
					dataChannel.onopen = resolve;
				});
				console.log("Channel opened!");
			}
			return {send,onMessage,close};
		});
	}
}